
////// fix for the problem that streams that need finishing (like release effects etc.) 
////// weren't able to be ended before their own substream ended.
////// (julian rohrhuber 6/2005)

////// the solution works as follows:

// a stream that wants to end its substreams must call the method .endStream.
// this will send a flag to all childstreams which, if they need finishing, should
// always check the inval for this flag. Then they add their cleanup events to a list called
// \endEvents which is passed on by endStream to be evaluated by the scheduler.
// the method copyToEndEvents is a convenience method to make adding events easier - it copies
// the current event into its own endEvents list.




//////////////////////////  streams finish other streams ////////////////////////



+ Pfin {
	embedInStream { arg event;
		var inevent;
		var stream = pattern.asStream;
		
		count.value.do {
			inevent = stream.next(event);
			if (inevent.isNil) { ^event };
			event = inevent.yield;
		};
		postf("Pfin % ending the stream\n", this.identityHash);
		
		^this.endStream(stream, event);
	}
}


+ Pfindur { // todo: swap event and inevent?
	embedInStream { arg event;
		var delta, elapsed = 0.0, nextElapsed, inevent, finishEvent;
	
		var stream = pattern.asStream;
		
		loop {
			inevent = stream.next(event);
			if(inevent.isNil) { ^event };
			delta = inevent.delta;
		//	postf("delta in Pfindur: % \n", delta);
			nextElapsed = elapsed + delta;
			if (nextElapsed.roundUp(tolerance) >= dur) {
				// must always copy an event before altering it.
				// fix delta time and yield to play the event.
				inevent = inevent.copy.put(\delta, dur - elapsed).yield;
				
				^this.endStream(stream, inevent);
			};

			elapsed = nextElapsed;
			event = inevent.yield;
		};
	}

}

// still to look at..

+ Pchain {
	embedInStream { arg inval, sawNil=false;
		var streams, inevent;
		streams = patterns.collect(_.asStream);
		loop {
			
			streams.reverseDo { |str|
				if(sawNil.not) {
					inevent = str.next(inval);
					if(inevent.isNil) { sawNil = true } { inval = inevent }
				}
			};
			
			// finish all streams
			if(sawNil or: {inval.eventAt(\type) == \end }) {
							streams.do { |stream|
								this.endStream(stream, inval);
							};
							^inval
			};
			if(sawNil) { ^inval };
			inval = yield(inevent);
		};
	}
}




//////////////////////////  streams that need finishing ////////////////////////

// Pmono + Pfin there is a problem: last event is used for switching off, instead
// of setting the synth one last time.

+ Pmono {

	embedInStream { arg inevent;
		var event, stream;
		var server, id;
		
		if(inevent.isNil) { ^inevent };
		
		server = inevent.eventAt(\server) ?? { Server.default };
		id = server.nextNodeID;
		
		// if(inevent.eventAt(\type) == \end) { inevent.put(\type, nil).yield; ^inevent };  
											// if early return avoid recursion
		
		if(inevent.eventAt(\type) == \end) { ^inevent.yield };								
		event = inevent.copy;
		event[\type] = \on;
		event[\delta] = 3e-10; // see below
		event[\id] = id;
		event[\instrument] = name;
		
		
		stream = Pbind(*([\type, \set, \args, args, \id, id] ++ patternpairs)).asStream;
		
//		server.makeSafeBundle(server.latency, {
//			inevent = event.yield; 
//			event = stream.next(inevent);
//			inevent = event.yield 
//		 });
// (we'll need some play thread local bundling mechanism.)
// the eventsStreamPlayer, passed down the stream, can have a \prepareEvents array.

		inevent = event.yield; 
		
		loop {
			event = stream.next(inevent);
			if(event.isNil or: {inevent.at(\type) == \end}) {
			//	postf("pmono received the end signal........ \n ....returning: % \n", inevent);
				event = inevent.copy;
				event[\id] = id;
				event[\delta] = 0;
				event[\type] = \off;
				inevent.addEndEvent(event);
				^inevent.yield
			};
			inevent = event.yield;
			
		}
	}
}



+ Pfx {

	embedInStream { arg inevent;	
		var stream;
		var server;
		var id;
		var event = inevent.copy;
		var finish = false, childHasFinished;
		
		
		// early finish
		if(inevent.isNil) { nil.yield; ^inevent }; // ### ??
		if(inevent.at(\type) == \end) { ^inevent.yield };								

		
		server = inevent[\server] ?? { Server.default };
		id = server.nextNodeID;
		
		// create and yield effect event
		pairs.pairsDo {|name, value|
			event[name] = value;
		};
		event[\addAction] = 1; // add to tail of group.
		event[\instrument] = fxname;
		event[\type] = \on;
		event[\id] = id;
		event[\delta] = 3e-10; // not good really.
		inevent = event.yield;
		
		stream = pattern.asStream;
		
		
		loop {
			event = stream.next(inevent);
			if(event.isNil or: {inevent.at(\type) == \end}) {
				event = inevent.copy;
				event[\id] = id;
				event[\delta] = 0;
				event[\type] = \off;
				^inevent.addEndEvent(event);
				//^inevent.yield;
			};
			inevent = event.yield;
			
		}	
	}

}


+ Ppar {
	embedInStream { arg inval;
		var assn;
		var priorityQ = PriorityQueue.new;
	
		repeats.value.do({ arg j;
			var outval, stream, nexttime, now = 0.0;

			this.initStreams(priorityQ);
			
			// if first event not at time zero
			if (priorityQ.notEmpty and: { (nexttime = priorityQ.topPriority) > 0.0 }, {
				outval = inval.copy;
				outval.put(\freq, \rest);					
				outval.put(\delta, nexttime);
				
				inval = outval.yield;
				now = nexttime;	
			});
			
			while({
				priorityQ.notEmpty
			},{
				// finish all streams
				if(inval.eventAt(\type) == \end) {
							while {priorityQ.notEmpty } {
								stream = priorityQ.pop;
								this.endStream(stream, inval);
							};
							^inval
				};
				
				stream = priorityQ.pop;
				outval = stream.next(inval);
				if (outval.isNil, {
					nexttime = priorityQ.topPriority;
					if (nexttime.notNil, {
						// that child stream ended, so rest until next one
						outval = inval.copy;
						outval.put(\freq, \rest);					
						outval.put(\delta, nexttime - now);
						
						inval = outval.yield;
						now = nexttime;	
					},{
						priorityQ.clear;
					});		
				},{			
					// requeue stream
					priorityQ.put(now + outval.delta, stream);
					nexttime = priorityQ.topPriority;
					outval.put(\delta, nexttime - now);
					
					inval = outval.yield;
					now = nexttime;	
				});	
			});
		});
		^inval;
	}
}	

// Pswitch1 can switch at any time.

+ Pswitch1 {
	
	embedInStream { arg inval;
		var index, outval;
		
		var streamList = list.collect({ arg pattern; pattern.asStream; });
		var indexStream = which.asStream;
		loop {
			// finish all streams
			if(inval.eventAt(\type) == \end) {
							streamList.do { |stream|
								this.endStream(stream, inval);
							};
							^inval
			};
			if ((index = indexStream.next).isNil) { ^inval };
			outval = streamList.wrapAt(index.asInteger).next(inval);
			if (outval.isNil) { ^inval };
			inval = outval.yield;
		};
	}

}




